import sys, subprocess
from termcolor import colored

is_windows = sys.platform == 'win32'
is_linux   = sys.platform.startswith('linux')
is_macos   = sys.platform == 'darwin'
if not (is_windows or is_linux or is_macos):
    raise RuntimeError(f'MamaBuild unsupported platform {sys.platform}')


def _is_system_64_bit():
    if sys.platform == 'win32':
        output = subprocess.check_output(['wmic', 'os', 'get', 'OSArchitecture'])
        if '64-bit' in str(output):
            return True
    else:
        output = subprocess.check_output(['uname', '-m'])
        if 'x86_64' in str(output):
            return True
    return False
is_64 = _is_system_64_bit()


class System:
    windows = is_windows
    linux   = is_linux
    macos   = is_macos
    is_64bit = is_64


class Color:
    DEFAULT = None
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"


# on windows use colorama to enable ANSI color escape sequences
if System.windows:
    from colorama import just_fix_windows_console
    just_fix_windows_console()


def get_colored_text(text:str, color):
    return colored(text, color=color) if color else text


def console(text:str, color=None, end="\n"):
    """ Always flush to support most build environments """
    print(get_colored_text(text, color), end=end, flush=True)


def error(text:str):
    """ Prints a message as an error, usually colored red """
    console(text, color=Color.RED)




import os
from .build_dependency import BuildDependency
from .util import read_lines_from, write_text_to
from .utils.system import console
import re

def write_default_mamafile(project_name, mamafile):
    contents = f'''import mama

##
# Explore Mama docs at https://github.com/RedFox20/Mama
#
class {project_name}(mama.BuildTarget):

    # this defines where to build all the dependencies
    # for project-local workspace: workspace = 'build'
    # for system-wide workspace: global_workspace = 'mycompany'
    workspace = 'build'

    # grab dependencies straight from git repositories
    # if the projects are trivial or support mama, then no extra configuration is needed
    # for others you will need to supply your own mamafile
    def dependencies(self):
        #self.add_git('ReCpp', 'https://github.com/RedFox20/ReCpp.git')
        #self.nothing_to_build() # if you have a header only library
        pass

    # customize CMake options in this step
    def configure(self):
        self.enable_cxx17()
        #self.add_cmake_options('BUILD_TESTS=ON', 'USE_SSE2=ON')

    ## optional: customize package exports if repository doesn't have `include` or `src`
    ##           default include and lib export works for most common static libs
    #def package(self):
    #    self.export_libs('.', ['.lib', '.a']) # export any .lib or .a from build folder
    #    self.export_includes(['include'])     # export 'include' path from source folder

    # run your custom testing steps here
    def test(self, args):
        self.gdb('bin/{project_name}', src_dir=True)

'''
    write_text_to(mamafile, contents)


def write_default_cmakelists(project_name, cmakefile):
    contents = f'''cmake_minimum_required(VERSION 3.8)
project({project_name})

# Include all mama dependencies via ${{MAMA_INCLUDES}} and ${{MAMA_LIBS}}
# For each dependency there will also be ${{SomeLibrary_LIBS}} (case-sensitive)
include(mama.cmake)
include_directories(${{MAMA_INCLUDES}})

# Executable {project_name}
include_directories("include")
file(GLOB_RECURSE PUBLIC_INTERFACE include/*.h)
file(GLOB_RECURSE PRIVATE_SOURCES  src/*.c  src/*.cpp  src/*.h)
source_group(include FILES ${{PUBLIC_INTERFACE}})
source_group(src     FILES ${{PRIVATE_SOURCES}})

add_executable({project_name} ${{PUBLIC_INTERFACE}} ${{PRIVATE_SOURCES}})
target_link_libraries({project_name} ${{MAMA_LIBS}})

install(FILES ${{PUBLIC_INTERFACE}} DESTINATION include)
install(TARGETS {project_name}
        RUNTIME DESTINATION bin
        LIBRARY DESTINATION lib)

'''
    write_text_to(cmakefile, contents)


def patch_existing_cmakelists(project_name, cmakefile):
    lines = read_lines_from(cmakefile)

    for i in range(len(lines)):
        if lines[i].startswith('include(mama.cmake)'):
            console(f'  Found include(mama.cmake) at line {i+1}. Not injecting basic mama includes.')
            return # Nothing to do!

    found_project = False
    for i in range(len(lines)):
        if bool(re.match('project\\s?\\(', lines[i], re.I)): # Match "project(" or "PROJECT (" 
            at = i+1
            lines[at:at] = [
                '\n',
                '# Include all mama dependencies via ${MAMA_INCLUDES} and ${MAMA_LIBS}\n',
                '# For each dependency there will also be ${SomeLibrary_LIBS} (case-sensitive)\n',
                'include(mama.cmake)\n',
                'include_directories(${MAMA_INCLUDES})\n',
            ]
            console(f'  Inserted include(mama.cmake) at line {at+1}.')
            found_project = True
            break
    if not found_project:
        console(f'  Could not find project() statement. Invalid CMakeLists.txt?')
        return

    inserted_link_lib = False
    for i in range(len(lines)):
        line = lines[i]
        if line.startswith('target_link_libraries('):
            if 'MAMA_LIBS' in line:
                console(f'  Already found ${{MAMA_LIBS}} at line {i+1}')
                inserted_link_lib = True
                break

            idx = line.find(')')
            if idx == -1:
                lines.insert(i+1, '    ${MAMA_LIBS}\n') # multiline target_line_libraries
                console(f'  Inserted ${{MAMA_LIBS}} at line {i+2}')
            else:
                lines[i] = line[:idx] + ' ${MAMA_LIBS}' + line[idx:] # add just before closing )
                console(f'  Inserted ${{MAMA_LIBS}} at line {i+1}')
            inserted_link_lib = True
            break

    if not inserted_link_lib:
        console(f'  Could not find suitable target_link_libraries() for ${{MAMA_LIBS}}. Please insert one manually to your CMakeLists.txt')

    contents = ''.join(lines)
    write_text_to(cmakefile, contents)


def mama_init_project(root: BuildDependency):
    mamafile = root.mamafile_path()
    if not os.path.exists(mamafile):
        console(f'{root.name} Creating new mamafile.py: {mamafile}')
        write_default_mamafile(root.name, root.mamafile_path())
    else:
        console(f'{root.name} Mamafile already exists: {mamafile}')

    cmakelists = root.cmakelists_path()
    if not os.path.exists(cmakelists):
        console(f'{root.name} Creating new CMakeLists.txt: {cmakelists}')
        write_default_cmakelists(root.name, cmakelists)
    else:
        console(f'{root.name} Patching existing CMakeLists.txt: {cmakelists}')
        patch_existing_cmakelists(root.name, cmakelists)





from __future__ import annotations
from typing import Tuple, TYPE_CHECKING
import os, shlex, shutil
from .system import System
from .sub_process import execute_echo
from ..util import normalized_path

if TYPE_CHECKING:
    from ..build_target import BuildTarget


def get_cwd_exe_args(target: BuildTarget, command: str, cwd='', root_dir='') -> Tuple[str, str, str]:
    """ Extracts the `cwd`, `exe` and `args` from a command string """
    shell_args = shlex.split(command)
    program = shell_args[0]
    args = ' '.join(shell_args[1:]) if shell_args else ''
    #print(f'get_cwd_exe_args: program={program} args={args} cwd={cwd} root_dir={root_dir}')

    # add or remove .exe extension
    if System.windows and target.windows and not program.endswith('.exe'):
        program += '.exe'
    if (System.linux or System.macos) and (not target.windows) and program.endswith('.exe'):
        program = program[:-4]

    if root_dir:
        # if root_dir is set, then command will be run relative to it
        # program: bin/app.exe
        # cwd: /path/to/root_dir/bin
        # exe: /path/to/root_dir/bin/app.exe
        cwd = os.path.join(root_dir, os.path.dirname(program))
        if program.startswith('/'):
            exe = program # already absolute
        else:
            exe = shutil.which(program) # is it a common executable?
            if not exe:
                exe = os.path.join(root_dir, program) # turn relative to absolute
        #print(f'ROOT cwd={cwd} exe={exe} args={args}')
    elif cwd:
        # if CWD is set, then command will be run in this dir
        # program: bin/app.exe
        # cwd: /path/to/project
        # exe: /path/to/project/bin/app.exe
        if program.startswith('/'):
            exe = program # already absolute
        else:
            exe = shutil.which(program) # is it a common executable?
            if not exe:
                exe = os.path.join(cwd, program) # turn relative to absolute
        #print(f'CWD cwd={cwd} exe={exe} args={args}')
    else:
        # otherwise the command will be run at the same dir as the executable
        # program: bin/app.exe
        # cwd: /path/to/bin
        # exe: /path/to/bin/app.exe
        cwd = os.path.dirname(os.path.abspath(program))
        # is it a common executable?
        exe = shutil.which(program)
        if not exe:
            exe = f'{cwd}/{os.path.basename(program)}'
        #print(f'DEFAULT cwd={cwd} exe={exe} args={args}')

    cwd = normalized_path(cwd)
    exe = normalized_path(exe)
    return cwd, exe, args


def run_in_working_dir(target: BuildTarget, working_dir: str, command: str, exit_on_fail=True):
    cwd, exe, args = get_cwd_exe_args(target, command, cwd=working_dir)
    execute_echo(cwd=cwd, cmd=f'{exe} {args}', exit_on_fail=exit_on_fail)


def run_in_project_dir(target: BuildTarget, command: str, src_dir=False, exit_on_fail=True):
    cwd = target.source_dir() if src_dir else target.build_dir()
    cwd, exe, args = get_cwd_exe_args(target, command, cwd=cwd)
    execute_echo(cwd=cwd, cmd=f'{exe} {args}', exit_on_fail=exit_on_fail)


def run_in_command_dir(target: BuildTarget, command: str, src_dir=False, exit_on_fail=True):
    root_dir = target.source_dir() if src_dir else target.build_dir()
    cwd, exe, args = get_cwd_exe_args(target, command, root_dir=root_dir)
    execute_echo(cwd=cwd, cmd=f'{exe} {args}', exit_on_fail=exit_on_fail)




from __future__ import annotations
from typing import TYPE_CHECKING
import os
import shlex
import shutil
import mama.util
import mama.utils.sub_process as proc
from mama.utils.system import console

if TYPE_CHECKING:
    from mama.build_target import BuildTarget

######################################################################################

class GnuProject:
    """ 
    Helper utility class for compiling GNU projects.
    This will enable downloading, unzipping and building GNU projects.

    Example usage:
    ```
    gmp = GnuProject(target, 'gmp', '6.2.1', 'https://gmplib.org/download/gmp/{{project}}.tar.xz', 'lib/libgmp.a')
    gmp.build()
    """
    def __init__(self, target:BuildTarget, name:str, version:str, 
                 build_product:str,
                 url:str='',
                 git:str='',
                 autogen=False,
                 configure='configure'):
        """
        - target: mama.BuildTarget building this GnuProject and collecting the build products
        - name: name of the project, eg 'gmp'
        - version: version of the project, eg '6.2.1'
        - build_product: the final product to build, eg 'lib/libgmp.a'
        - url: url to download the project, eg 'https://gmplib.org/download/gmp/{{project}}.tar.xz'
        - git: git to clone the project from
        - autogen: whether to use ./autogen.sh before running ./configure
        - configure: the configuration command, by default 'configure' but can be 'make config' etc
        """
        self.target = target
        self.name = name
        self.version = version
        self.name_with_version = f'{name}-{version}'
        self.url = url
        self.git = git
        self.autogen = autogen
        self.build_product = build_product
        self.install_dir_suffix = '-built'
        self.make_opts = '' # default options for make
        self.host = ''
        if self.target.config.mips:
            self.host = 'mipsel-elf'
        # the configure command, by default it's 'configure'
        # however using something other than 'configure' will completely override it
        self.configure_command = configure


    def source_dir(self, subpath=''):
        """ Where the project is extracted to, eg project/mips/gdb-13.2 """
        if subpath:
            return self.target.build_dir(self.name_with_version + '/' + subpath)
        return self.target.build_dir(self.name_with_version)


    def install_dir(self, subpath=''):
        """ Where the project is installed to, eg project/mips/gdb-built """
        if subpath:
            return self.target.build_dir(self.name + self.install_dir_suffix + '/' + subpath)
        return self.target.build_dir(self.name + self.install_dir_suffix)


    def get_final_product(self):
        """ Returns the final product path, eg project/mips/gdb-built/bin/gdb """
        product = self.build_product
        if '{{source}}' in product:
            return product.replace('{{source}}', self.source_dir())
        if '{{build}}' in product:
            return product.replace('{{build}}', self.target.build_dir())
        return self.install_dir() + "/" + product


    def should_build(self):
        """ Returns true if the final product does not exist """
        return not os.path.exists(self.get_final_product())


    def get_makefile(self):
        """ Gets the Makefile, which is a build step dependency """
        return f'{self.source_dir()}/Makefile'


    def get_configure_file(self):
        """ Gets the file which performs configuration, and is thus a configuration step dependency """
        args = shlex.split(self.configure_command)
        config_cmd = args[0].strip()
        if config_cmd == 'make':  # if config_cmd is 'make', then look for a Makefile
            return self.get_makefile()
        return f'{self.source_dir()}/{config_cmd}'


    def checkout_code(self):
        """
        Checks out the code archive, either by downloading and extracting the zip archive, 
        or cloning the project from git repository.
        """
        source = self.source_dir()
        configure_file = self.get_configure_file()
        if os.path.exists(configure_file):
            return # nothing to do

        build_root = self.target.build_dir()
        if self.git:
            console(f'>>> Cloning {source} from {self.git}', color='green')
            if os.system(f'git clone {self.git} {source}') != 0:
                raise Exception(f'Failed to clone {self.git} to {source}')
        else:
            url = self.url.replace('{{project}}', self.name_with_version)
            local_file = mama.util.download_file(url, local_dir=build_root)

            console(f'>>> Extracting to {source}', color='green')
            os.makedirs(source, exist_ok=True)
            if local_file.endswith('.tar.xz') \
                or local_file.endswith('.tar.gz') \
                or local_file.endswith('.tar.gz2') \
                or local_file.endswith('.tar.bz2'):
                proc.execute_echo(build_root, f'tar -xf {local_file} -C {source} --strip-components=1')
            elif local_file.endswith('.zip'):
                mama.util.unzip(local_file, source)
            else:
                console(f'>>> ERROR: Unknown archive type: {local_file}', color='red')

        # final check if the configure file exists
        if not os.path.exists(configure_file):
            raise Exception(f'Checkout failed, no configure file at: {configure_file}')


    def configure_env(self):
        # GNU projects need to be configured with the CC, CXX and AR environment variables set
        cc_prefix = self.target.get_cc_prefix()
        os.environ['CC'] = cc_prefix + 'gcc'
        os.environ['CXX'] = cc_prefix + 'g++'
        os.environ['AR'] = cc_prefix + 'ar'
        os.environ['LD'] = cc_prefix + 'ld'
        os.environ['READELF'] = cc_prefix + 'readelf'
        os.environ['STRIP'] = cc_prefix + 'strip'
        os.environ['RANLIB'] = cc_prefix + 'ranlib'


    def run(self, command):
        """ Runs a command in the project directory, eg 'make specialsetup' """
        self.target.run_program(self.source_dir(), command)


    def configure(self, options='', prefix=''):
        """
        Only configures the project for building by generating the Makefile
            - options: additional options to pass to the configure script
        """
        console(f'>>> Configuring {self.name}', color='green')
        self.configure_env()

        if self.autogen:
            self.run('./autogen.sh')

        if self.configure_command != 'configure':
            # user overrides with custom configurator, such as `make config` etc
            configure = self.configure_command
        else:
            args = f'--host={self.host}'
            if not prefix:
                prefix = f'--prefix {self.install_dir()}'

            if not self.autogen:
                guess_machine = f'{self.source_dir()}/config.guess'
                if os.path.exists(guess_machine):
                    args += f' --build={proc.execute_piped(guess_machine)}'

            configure = f'./configure {args} {prefix}'

        self.run(f'{configure} {options}')
        console(f'>>> Configured {self.name}', color='green')


    def _get_make_opts(self, opts, multithreaded=False):
        jobs = f'-j {self.target.config.jobs}' if multithreaded else ''
        all_opts = ''
        for o in [self.make_opts, opts, jobs]:
            if all_opts: all_opts += ' '
            all_opts += o
        return all_opts


    def make(self, opts='', multithreaded=False):
        """
        Only makes the project
            - opts: extra options for make
            - multithreaded: if true, will use the -j option to build with multiple threads
        """
        make_opts = self._get_make_opts(opts, multithreaded)
        console(f'>>> Make {self.name} {make_opts}', color='green')
        self.configure_env()
        self.run(f'make {make_opts}')
        console(f'>>> Made {self.name} {make_opts}', color='green')


    def install(self, no_prefix=False):
        """ Only installs the project """
        console(f'>>> Installing {self.name}', color='green')
        self.configure_env()
        prefix = '' if no_prefix else f'PREFIX={self.install_dir()}'
        all_opts = self._get_make_opts(prefix, multithreaded=False)
        self.run(f'make {all_opts} install')
        console(f'>>> Installed {self.name}', color='green')


    def build(self, options='', make_opts='', prefix='',
              multithreaded=False, install=True):
        """
        Downloads, Unzips, Configures, Makes and Installs the project 
            - options: additional options to pass to the configure script
            - multithreaded: if true, will use the -j option to build with multiple threads
        """
        project_dir = self.source_dir()
        autoconf_makefile = f'{project_dir}/Makefile' if self.configure_command == 'configure' else None
        try:
            console(f'>>>>>> BUILD {self.name} <<<<<<', color='green')
            self.checkout_code()

            if not autoconf_makefile: # if not autoconf then always run the configure step
                self.configure(options=options, prefix=prefix)
            elif not os.path.exists(autoconf_makefile):
                self.configure(options=options, prefix=prefix)

            self.make(opts=make_opts, multithreaded=multithreaded)
            if install:
                self.install()
        except:
            console(f'>>> ERROR: Failed to build {self.name}', color='red')
            # with autoconf projects, delete the makefile so that it will be regenerated
            if autoconf_makefile and os.path.exists(autoconf_makefile):
                os.remove(autoconf_makefile)
            raise


    def strip(self, src_path, dest_path=None):
        striptool = self.target.get_cc_prefix() + 'strip'
        out = f'-o {dest_path}' if dest_path else ''
        if os.system(f'{striptool} {src_path} {out}') != 0:
            raise Exception(f'Failed to strip {src_path}')


    def copy_file_or_link(self, src_file, dst_file):
        """ Copies a file or symlink preserving their attributes and relative symlinks """
        if os.path.islink(src_file):
            link = os.readlink(src_file)
            #console(f'link: {dst_file} -> {link}', color='yellow')
            os.remove(dst_file)
            os.symlink(link, dst_file)
        else:
            mama.util.copy_file(src_file, dst_file)


    def deploy(self, src_path=None, dest_path=None, strip=False):
        """ 
        Deploys the src_path to dest_path, 
        optionally stripping the binary in the process.
        If `src_path` is None, the get_final_product() will be used
        """
        if not src_path:
            src_path = self.get_final_product()
        if not dest_path:
            dest_path = src_path
        if strip and os.path.isfile(src_path):
            self.strip(src_path, src_path)
        else:
            self.copy_file_or_link(src_path, dest_path)


    def deploy_dir(self, src_dir, dest_dir, strip=False):
        """ Deploys all built products from src_dir to dest_dir and optionally strips them """
        root = os.path.dirname(src_dir)
        for fulldir, _, files in os.walk(src_dir):
            for file in files:
                reldir = fulldir[len(root):].lstrip('\\/')
                if reldir:
                    dst_folder = os.path.join(dest_dir, reldir)
                else:
                    dst_folder = dest_dir
                os.makedirs(dst_folder, exist_ok=True)
                src_file = os.path.join(fulldir, file)
                dst_file = os.path.join(dst_folder, file)
                if strip and os.path.isfile(src_file):
                    self.strip(src_file, dest_path=dst_file)
                else:
                    self.copy_file_or_link(src_file, dst_file)


######################################################################################




from __future__ import annotations
from typing import TYPE_CHECKING, Callable
import os
from mama.utils.system import System, console

if TYPE_CHECKING:
    from ..build_config import BuildConfig
    from ..build_target import BuildTarget

class Android:
    def __init__(self, config: BuildConfig):
        ## Android NDK Clang
        self.config = config
        self.toolchain_file = None  ## possibility to override the toolchain file
        self.android_sdk_path = ''
        self.android_ndk_path = ''
        self.android_api = 'android-29' # 29: Android 10.0 (2020)
        self.android_ndk_stl = 'c++_shared' # LLVM libc++
        self.ndk_version = 'ndk'


    def android_abi(self):
        if self.config.is_target_arch_armv7(): return 'armeabi-v7a'
        elif self.config.arch == 'arm64': return 'arm64-v8a'
        else: raise RuntimeError(f'Unrecognized android arch: {self.config.arch}')


    def android_home(self):
        if not self.android_sdk_path: self.init_ndk_path()
        return self.android_sdk_path


    def android_ndk(self):
        if not self.android_ndk_path: self.init_ndk_path()
        return self.android_ndk_path


    def bin(self):
        if not self.android_ndk_path: self.init_ndk_path()
        platform_dir = 'linux-x86_64'
        if System.windows: platform_dir = 'windows-x86_64'
        return f'{self.android_ndk_path}/toolchains/llvm/prebuilt/{platform_dir}/bin'


    def cc_path(self):
        platform_ver = self.android_api.replace('-', '')
        arch = 'aarch64'
        if self.config.arch == 'arm64': arch = 'aarch64'
        elif self.config.arch == 'arm': arch = 'armv7a'
        elif self.config.arch == 'x64': arch = 'x86_64'
        elif self.config.arch == 'x86': arch = 'i686'
        return f'{self.bin()}/{arch}-linux-{platform_ver}-clang'


    def cxx_path(self):
        return f'{self.cc_path()}++'


    def set_toolchain_path(self, toolchain_file: str):
        if not os.path.exists(toolchain_file):
            raise RuntimeError(f'Android toolchain file not found: {toolchain_file}')
        self.toolchain_file = toolchain_file


    def _set_ndk_sdk_paths(self, ndk_path: str, sdk_path: str):
        self.android_sdk_path = sdk_path
        self.android_ndk_path = ndk_path
        if self.config.print:
            console(f'Found Android SDK: {self.android_sdk_path}')
            console(f'Found Android NDK: {self.android_ndk_path}')


    @staticmethod
    def _append_env(paths:list, env: str):
        path = os.getenv(env)
        if path and os.path.exists(path):
            paths.append(path)


    def init_ndk_path(self):
        ndk_build = 'ndk-build.cmd' if System.windows else 'ndk-build'

        # prefer NDK paths to find SDK and NDK
        ndk_paths = []
        Android._append_env(ndk_paths, 'ANDROID_NDK_LATEST_HOME')
        Android._append_env(ndk_paths, 'ANDROID_NDK_HOME')
        Android._append_env(ndk_paths, 'ANDROID_NDK_ROOT')
        Android._append_env(ndk_paths, 'ANDROID_NDK')

        for ndk_path in ndk_paths:
            if ndk_path and os.path.exists(f'{ndk_path}/{ndk_build}'):
                # figure out the Sdk path
                if os.path.exists(f'{ndk_path}/../../platforms'):
                    self._set_ndk_sdk_paths(ndk_path, os.path.abspath(f'{ndk_path}/../..'))
                    return
                elif os.path.exists(f'{ndk_path}/../platforms'):
                    self._set_ndk_sdk_paths(ndk_path, os.path.abspath(f'{ndk_path}/..'))
                    return
                else:
                    continue # try next path

        # find SDK and NDK by using SDK root
        sdk_paths = []
        Android._append_env(sdk_paths, 'ANDROID_HOME')
        Android._append_env(sdk_paths, 'ANDROID_SDK_ROOT')

        if System.windows:
            sdk_paths += [f'{os.getenv("LOCALAPPDATA")}\\Android\\Sdk']
        elif System.linux:
            sdk_paths += [
                f'{os.getenv("HOME")}/Android/Sdk',
                '/usr/bin/android-sdk',
                '/opt/android-sdk',
                '/opt/Android',
                '/Android'
            ]
        elif System.macos:
            sdk_paths += [f'{os.getenv("HOME")}/Library/Android/sdk']

        for sdk_path in sdk_paths:
            # older NDK versions with ndk-bundle subdir
            if os.path.exists(f'{sdk_path}/ndk-bundle/{ndk_build}'):
                self._set_ndk_sdk_paths(sdk_path + '/ndk-bundle', sdk_path)
                return
            # newer NDK with multiple ndk versions:
            elif os.path.exists(f'{sdk_path}/ndk'):
                subdirs = os.listdir(f'{sdk_path}/ndk')
                subdirs.sort(reverse=True)
                for subdir in subdirs:
                    if os.path.exists(f'{sdk_path}/ndk/{subdir}/{ndk_build}'):
                        self.ndk_version = subdir
                        self._set_ndk_sdk_paths(f'{sdk_path}/ndk/{subdir}', sdk_path)
                        return
        raise EnvironmentError(f'''Could not detect any Android NDK installations. 
Default search paths: {ndk_paths+sdk_paths} 
Define env ANDROID_NDK_HOME with path to the preferred NDK installation
Or define env ANDROID_HOME with path to Android SDK root with valid NDK-s.''')


    def get_cxx_flags(self, add_flag: Callable[[str,str], None]):
        if self.config.is_target_arch_armv7():
            add_flag('-march', 'armv7-a')
            add_flag('-mfpu', 'neon')
        else:
            add_flag('-march', 'armv8-a')

        # this was only needed for legacy NDK toolchains
        # but on latest toolchains these issues are fixed
        # add_flag(f'-I"{self.android_ndk()}/toolchains/llvm/prebuilt/linux-x86_64/sysroot/usr/include"')
        # add_flag(f'-I"{self.android_ndk()}/sources/cxx-stl/llvm-libc++/include"')


    def _get_make(self):
        if System.windows:
            return f'{self.android_ndk()}\\prebuilt\\windows-x86_64\\bin\\make.exe' # CodeBlocks - Unix Makefiles
        elif System.macos:
            return f'{self.android_ndk()}/prebuilt/darwin-x86_64/bin/make' # CodeBlocks - Unix Makefiles
        return ''


    def _get_toolchain_path(self, target: BuildTarget) -> str:
        if target.cmake_ndk_toolchain:
            if os.path.isabs(target.cmake_ndk_toolchain):
                toolchain = target.cmake_ndk_toolchain
            else:
                toolchain = target.source_dir(target.cmake_ndk_toolchain)
            if os.path.exists(toolchain):
                return toolchain

        if self.toolchain_file and os.path.exists(self.toolchain_file):
            return self.toolchain_file

        toolchain = f'{self.android_ndk()}/build/cmake/android.toolchain.cmake'
        if os.path.exists(toolchain):
            return toolchain
        return ''


    def get_cmake_build_opts(self, target: BuildTarget) -> list:
        opts = [
            'CMAKE_SYSTEM_NAME=Android',
            f'ANDROID_ABI={self.android_abi()}',
            'ANDROID_ARM_NEON=TRUE',
            f'ANDROID_NDK="{self.android_ndk()}"',
            f'ANDROID_STL={self.android_ndk_stl}',
            f'ANDROID_NATIVE_API_LEVEL={self.android_api}',
            'ANDROID_TOOLCHAIN=clang',
            'CMAKE_BUILD_WITH_INSTALL_RPATH=ON',
            'ANDROID_USE_LEGACY_TOOLCHAIN_FILE=FALSE'
        ]

        # get the toolchain file overriden by build target
        target_toolchain = target.source_dir(target.cmake_ndk_toolchain)
        if target.cmake_ndk_toolchain and os.path.exists(target_toolchain):
            toolchain = target_toolchain
        elif self.toolchain_file and os.path.exists(self.toolchain_file):
            toolchain = self.toolchain_file
        else:
            toolchain = f'{self.android_ndk()}/build/cmake/android.toolchain.cmake'

        if toolchain:
            opts.append(f'CMAKE_TOOLCHAIN_FILE="{toolchain}"')
            if self.config.print:
                console(f'Toolchain: {toolchain}')

        make = self._get_make()
        if make: opts.append(f'CMAKE_MAKE_PROGRAM="{make}"')
        return opts

    # injects android specific env vars
    def inject_env(self):
        make = self._get_make()
        if make: os.environ['CMAKE_MAKE_PROGRAM'] = make
        os.environ['ANDROID_HOME'] = self.android_home()
        os.environ['ANDROID_NDK'] = self.android_ndk()
        os.environ['ANDROID_ABI'] = self.android_abi()
        os.environ['ANDROID_STL'] = self.android_ndk_stl
        os.environ['ANDROID_NATIVE_API_LEVEL'] = self.android_api
        os.environ['ANDROID_TOOLCHAIN'] = 'clang'



import os, sys, tempfile, platform, psutil, shutil
from typing import List
from mama.platforms.oclea import Oclea
from mama.platforms.mips import Mips
from mama.platforms.android import Android
import mama.util as util
from .utils.system import System, console
from .utils.sub_process import execute, execute_piped

if System.linux:
    import distro


###
# Mama Build Configuration is created only once in the root project working directory
# This configuration is then passed down to dependencies
#
class BuildConfig:
    def __init__(self, args):
        # commands
        self.list    = False
        self.build   = False
        self.clean   = False
        self.rebuild = False
        self.update  = False
        self.deploy  = False
        # if root mamafile has defined an artifacts URL
        # this will upload deploy archive through SFTP
        self.upload  = False
        # currently for uploads only, uploads only if package already not uploaded
        self.if_needed = False
        # if `art` is specified, then artifactory download is mandatory, no source builds are done
        self.force_artifactory = False
        # if `noart` is specified, then artifactory is temporarily ignored
        self.disable_artifactory = False
        self.reclone   = False
        self.mama_init = False
        self.print     = True
        self.verbose   = False
        self.test      = ''
        self.start     = ''
        # supported platforms
        self.windows = False
        self.linux   = False
        self.macos   = False
        self.ios     = False
        self.android : Android = None
        self.raspi   = False
        self.oclea : Oclea = None
        self.mips : Mips = None
        # compilers
        self.clang = True # prefer clang on linux
        self.gcc   = False
        self.clang_path = ''
        self.gcc_path = ''
        # can be used to overide C and C++ compiler paths
        self.cc_path = ''
        self.cxx_path = ''
        self.cxx_version = '' # c++ compiler version, eg '8.3.0' for gcc 8.3.0
        # If compiler specificed from command line
        # using `mama build gcc` or `mama build clang`
        self.compiler_cmd = False
        self.fortran = ''
        # build optimization
        self.release = True
        self.debug   = False
        # valid architectures: x86, x64, arm, arm64
        self.arch    = None
        self.distro  = None  # distro information (name, major, minor)
        self.jobs    = psutil.cpu_count()
        self.target  = None
        self.flags   = None
        self.open    = None
        # use this to customize ios sdk version
        self.ios_version   = '16.0' # 16: ios 16
        # use this to customize macos sdk version
        self.macos_version = '13.0' # 13: macos 13
        ## Artifactory URL for dependency uploads and downloads
        self.artifactory_ftp = None
        self.artifactory_auth = None
        ## Ninja
        self.ninja_path = self.find_ninja_build()
        ## MSVC, MSBuild
        self._visualstudio_path = None
        self._visualstudio_cmake_id = None
        self._msbuild_path = None
        self._msvctools_path = None
        ## Raspberry PI - Raspi
        self.raspi_compilers  = ''  ## Raspberry g++ and gcc
        self.raspi_system     = ''  ## path to Raspberry system libraries
        self.raspi_include_paths = [] ## path to additional Raspberry include dirs
        ## Convenient installation utils:
        self.convenient_install = []
        ## Workspace and parsing
        self.parallel_load = False  ## Whether to load dependencies in parallel?
        self.global_workspace = False
        if System.windows:
            self.workspaces_root = util.normalized_path(os.getenv('HOMEPATH'))
        else:
            self.workspaces_root = os.getenv('HOME')
        self.unused_args = []
        self.parse_args(args)
        self.check_platform()


    def parse_args(self, args: List[str]):
        for arg in args:
            if   arg == 'list':      self.list    = True
            elif arg == 'build':     self.build   = True
            elif arg == 'clean':     self.clean   = True
            elif arg == 'rebuild':   self.rebuild = True
            elif arg == 'update':    self.update  = True
            elif arg == 'deploy':    self.deploy  = True
            elif arg == 'upload':    self.upload  = True
            elif arg == 'if_needed': self.if_needed = True
            elif arg == 'art':       self.force_artifactory = True
            elif arg == 'noart':     self.disable_artifactory = True
            # Updates, Builds and Deploys the project as a package
            elif arg == 'serve':
                self.rebuild = True
                self.update = True
                self.deploy = True
                self.upload = True
            elif arg == 'reclone':
                console('WARNING: Argument `reclone` is deprecated, use `wipe` instead.')
                self.reclone = True
            elif arg == 'wipe':      self.reclone = True
            elif arg == 'init':      self.mama_init = True
            elif arg == 'silent':    self.print = False
            elif arg == 'verbose':   self.verbose = True
            elif arg == 'parallel':  self.parallel_load = True
            elif arg == 'all':       self.target = 'all'
            elif arg == 'test':      self.test = ' ' # no test arguments
            elif arg == 'start':     self.start = ' ' # no start arguments
            elif arg == 'windows': self.set_platform(windows=True)
            elif arg == 'linux':   self.set_platform(linux=True)
            elif arg == 'macos':   self.set_platform(macos=True)
            elif arg == 'ios':     self.set_platform(ios=True)
            elif arg == 'android': self.set_platform(android=True)
            elif arg == 'raspi':   self.set_platform(raspi=True)
            elif arg == 'oclea':   self.set_platform(oclea=True)
            elif arg == 'mips':    self.set_platform(mips=True)
            elif arg == 'x86':     self.set_arch('x86')
            elif arg == 'x64':     self.set_arch('x64')
            elif arg == 'arm':     self.set_arch('arm')
            elif arg == 'arm64':   self.set_arch('arm64')
            elif arg == 'aarch64': 
                console('warning: aarch64 is the same as arm64, setting to arm64')
                self.set_arch('arm64')
            elif arg == 'clang':
                self.gcc = False
                self.clang = True
                self.compiler_cmd = True
            elif arg == 'gcc':
                self.gcc = True
                self.clang = False
                self.compiler_cmd = True
            elif arg == 'fortran': self.fortran = self.find_default_fortran_compiler()
            elif arg == 'fortran=': self.fortran = arg[8:]
            elif arg == 'release': self.set_build_config(release=True)
            elif arg == 'debug':   self.set_build_config(debug=True)
            elif arg == 'open':    self.open = 'root'
            # Open a specific target source dir for editing with VSCode or Visual Studio
            # Ex old: mama open=ReCpp
            # Ex new: mama open ReCpp
            elif arg.startswith('open='):   self.open = arg[5:]
            elif arg.startswith('jobs='):   self.jobs = int(arg[5:])
            # Sets the target to build/update/clean
            # This is superceded by automatic target lookup
            # Ex old: mama build target=opencv
            # Ex new: mama build opencv
            elif arg.startswith('target='): self.target = arg[7:]
            # Adding arguments for tests runner
            # Ex: mama build test="nogdb threadpool"
            # Ex: mama build test=nogdb test=threadpool
            # Ex: mama build test=nogdb,threadpool
            elif arg.startswith('test='):   self.test = self.join_args(self.test, arg[5:])
            # Calls target.start with the specified arguments
            # Ex: mama build start=verify
            elif arg.startswith('start='):  self.start = self.join_args(self.start, arg[6:])
            elif arg.startswith('arch='):   self.set_arch(arg[5:])
            # Add additional compiler flags
            elif arg.startswith('flags='):  self.flags = self.join_args(self.flags, arg[6:])
            # Ex: mama build android-24
            elif arg.startswith('android-'):
                self.set_platform(android=True)
                self.android.android_api = arg
            elif arg == 'install-clang6':  self.convenient_install.append('clang6')
            elif arg == 'install-clang11': self.convenient_install.append('clang11')
            elif arg == 'install-msbuild': self.convenient_install.append('msbuild')
            elif arg == 'install-ndk':     self.convenient_install.append('ndk')
            else:
                self.unused_args.append(arg)
            continue


    # modifies existing `args` by parsing and appending `arg` contents
    def join_args(self, args, arg):
        if arg[0] == '"' and arg[-1] == '"':
            arg = arg[1:-1]
        elif ',' in arg:
            arg = ' '.join(arg.split(','))
        if not args:
            return arg
        return args + ' ' + arg


    def set_platform(self, windows=False, linux=False, macos=False, \
                           ios=False, android=False, raspi=False, \
                           oclea=False, mips=False):
        """ Ensures only a single platform is set """
        platforms = [False]*8
        if windows: platforms[0] = True
        elif linux:  platforms[1] = True
        elif macos:  platforms[2] = True
        elif ios:    platforms[3] = True
        elif android: platforms[4] = True
        elif raspi: platforms[5] = True
        elif oclea: platforms[6] = True
        elif mips: platforms[7] = True

        def get_new_value(old_value, enable, type=None):
            if old_value and not enable:
                return None if type else False
            if enable and not old_value:
                return type(self) if type else True
            return old_value
        self.windows = get_new_value(self.windows, platforms[0])
        self.linux   = get_new_value(self.linux,   platforms[1])
        self.macos   = get_new_value(self.macos,   platforms[2])
        self.ios     = get_new_value(self.ios,     platforms[3])
        self.android = get_new_value(self.android, platforms[4], Android)
        self.raspi   = get_new_value(self.raspi,   platforms[5])
        self.oclea   = get_new_value(self.oclea,   platforms[6], Oclea)
        self.mips    = get_new_value(self.mips,    platforms[7], Mips)
        return True


    def is_platform_set(self):
        return self.windows or self.linux or self.macos \
            or self.ios or self.android or self.raspi \
            or self.oclea or self.mips


    def check_platform(self):
        if not self.is_platform_set():
            self.set_platform(windows=System.windows, linux=System.linux, macos=System.macos)
            if not self.is_platform_set():
                raise RuntimeError(f'Unsupported platform {sys.platform}: Please specify platform!')

        # set defaults if arch was not specified
        if not self.arch:
            # macos has now moved to arm64 starting with M1 series chips
            if self.macos:        self.set_arch('arm64')
            elif self.ios:        self.set_arch('arm64')
            elif self.android:    self.set_arch('arm64')
            elif self.raspi:      self.set_arch('arm')
            elif self.oclea:      self.set_arch('arm64')
            elif self.mips:       self.set_arch(self.mips.mips_arch)
            elif System.is_64bit: self.set_arch('x64')
            else:                 self.set_arch('x86')

        # Arch itself is validated in set_arch(), 
        # however we need to validate if arch is allowed on platform
        if self.arch:
            if self.linux and 'arm' in self.arch:
                raise RuntimeError(f'Unsupported arch={self.arch} on linux platform! Build with android instead')
            if self.raspi and self.arch != 'arm':
                raise RuntimeError(f'Unsupported arch={self.arch} on raspi platform! Supported=arm')
            if self.oclea and self.arch != 'arm64':
                raise RuntimeError(f'Unsupported arch={self.arch} on Oclea platform! Supported=arm64')
            if self.mips and self.arch not in self.mips.supported_arches:
                raise RuntimeError(f'Unsupported arch={self.arch} on MIPS platform! Supported={self.mips.supported_arches}')


    def get_distro_info(self):
        if self.distro:
            return self.distro
        if self.windows:
            version = platform.version().split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.macos:
            version = self.macos_version.split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.ios:
            version = self.ios_version.split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.android:
            version = self.android.android_api.split('-')[1]
            self.distro = (self.name(), int(version), 0)
        elif self.raspi:
            # TODO: RASPI version
            self.distro = (self.name(), 0, 0)
        elif self.oclea:
            # TODO: OCLEA version
            self.distro = (self.name(), 0, 0)
        elif self.mips:
            # TODO: MIPS version
            self.distro = (self.name(), self.mips.toolchain_major, self.mips.toolchain_minor)
        elif self.linux:
            try:
                dist = distro.info()
                major = int(dist['version_parts']['major'])
                minor = int(dist['version_parts']['minor'])
                self.distro = (dist['id'], major, minor)
            except Exception as err:
                console(f'Failed to parse linux distro; falling back to Ubuntu 16.04 LTS: {err}', color='red')
                self.distro = ('ubuntu', 16, 4)
        else:
            self.distro = (platform.system().lower(), int(platform.release()), 0)
        return self.distro


    def set_arch(self, arch):
        arches = ['x86', 'x64', 'arm', 'arm64', 'mips', 'mipsel', 'mips64', 'mips64el']
        if not arch in arches:
            raise RuntimeError(f"Unrecognized architecture {arch}! Valid options are: {arches}")
        self.arch = arch


    def is_64bit_build(self):
        return (self.arch == 'x64' or self.arch == 'arm64')


    def name(self):
        if self.windows: return 'windows'
        if self.linux:   return 'linux'
        if self.macos:   return 'macos'
        if self.ios:     return 'ios'
        if self.android: return 'android'
        if self.raspi:   return 'raspi'
        if self.oclea:   return 'oclea'
        if self.mips:    return self.mips.name
        return 'build'


    ## These are the hard references to all build directory variations
    ## All parts of the codebase should use these, instead of raw strings
    ## This will avoid accidental mismatches
    def build_dir_win64(self): return 'windows'
    def build_dir_win32(self): return 'windows32'
    def build_dir_winarm64(self): return 'winarm'
    def build_dir_winarm32(self): return 'winarm32'
    def build_dir_linux64(self): return 'linux'
    def build_dir_linux32(self): return 'linux32'
    def build_dir_macosarm64(self): return 'macosarm' # arm64
    def build_dir_macos64(self): return 'macos' # x64
    def build_dir_ios(self): return 'ios' # arm64
    def build_dir_android64(self): return 'android'
    def build_dir_android32(self): return 'android32'
    def build_dir_raspi32(self): return 'raspi'
    def build_dir_oclea64(self): return 'oclea'
    def build_dir_mips(self): return 'mips'
    def build_dir_default(self): return 'build'


    def platform_build_dir_name(self):
        """
        Gets the build folder name depending on platform and architecture.
        By default 64-bit architectures use the platform name, eg 'windows' or 'linux'
        And 32-bit architectures add a suffix, eg 'windows32' or 'linux32'
        """
        # WARNING: This needs to be in sync with dependency_chain.py: _save_mama_cmake !!!
        if self.windows:
            if self.is_target_arch_x64(): return self.build_dir_win64()
            if self.is_target_arch_x86(): return self.build_dir_win32()
            if self.is_target_arch_armv7(): return self.build_dir_winarm32()
            return self.build_dir_winarm64()
        if self.linux:
            if self.is_target_arch_x64(): return self.build_dir_linux64()
            return self.build_dir_linux32()
        if self.macos: # Apple dropped 32-bit support
            # and new default should be arm64 starting from M1 series chips
            if self.is_target_arch_arm64(): return self.build_dir_macosarm64()
            return self.build_dir_macos64()
        if self.ios: # Apple dropped 32-bit support
            return self.build_dir_ios()
        if self.android:
            if self.is_target_arch_arm64(): return self.build_dir_android64()
            return self.build_dir_android32()
        if self.raspi: return self.build_dir_raspi32()  # Only 32-bit raspi
        if self.oclea: return self.build_dir_oclea64()  # Only 64-bit oclea aarch64 (arm64)
        if self.mips: return self.build_dir_mips()

        return self.build_dir_default()


    def set_build_config(self, release=False, debug=False):
        self.release = release
        self.debug   = debug
        return True


    def set_artifactory_ftp(self, ftp_url: str, auth='store'):
        """ @see BuildTarget.set_artifactory_ftp() for documentation """
        self.artifactory_ftp = ftp_url
        self.artifactory_auth = auth


    def prefer_clang(self, target_name):
        if not self.linux or self.raspi or self.clang: return
        if not self.compiler_cmd:
            self.clang = True
            self.gcc   = False
            self.compiler_cmd = True
            if self.print:
                console(f'Target {target_name} requests Clang. Using Clang since no explicit compiler flag passed.')
        else:
            if self.print: 
                console(f'Target {target_name} requested Clang but compiler already set to GCC.')


    def prefer_gcc(self, target_name):
        if not self.linux or self.raspi or self.gcc: return
        if not self.compiler_cmd:
            self.clang = False
            self.gcc   = True
            self.compiler_cmd = True
            if self.print:
                console(f'Target {target_name} requests GCC. Using GCC since no explicit compiler flag passed.')
        else:
            if self.print:
                console(f'Target {target_name} requested GCC but compiler already set to Clang.')


    ##
    # Enables fortran compiler
    # @path Optional custom path or command for the Fortran compiler 
    #
    def enable_fortran(self, path=''):
        if self.fortran: return
        self.fortran = path if path else self.find_default_fortran_compiler()


    # returns: root path where the compilers exist and the discovered suffix
    #          (root_path, suffix)
    def find_compiler_root(self, suggested_path, compiler, suffixes, dumpfullversion):
        roots = []
        if suggested_path: roots.append(suggested_path)
        roots += ['/etc/alternatives/', '/usr/bin/', '/usr/local/bin/', '/bin/']
        candidates = []
        for root in roots:
            for suffix in suffixes:
                cc_path = root + compiler + suffix
                if os.path.exists(cc_path):
                    version = self._get_gcc_clang_fullversion(cc_path, dumpfullversion) # eg 9.4.0
                    if self.verbose: console(f'Compiler {cc_path} version: {version}')
                    candidates.append((root, suffix, version))
        if not candidates:
            raise EnvironmentError(f'Could not find {compiler} from {roots} with any suffix {suffixes}')

        def version_to_int(version_str):
            major_minor_patch = version_str.split('.')
            integer = 0
            for part in major_minor_patch:
                integer = integer*10 + int(part) if part else integer
            if integer == 0:
                console(f"Failed to check version for candidate='{version_str}'")
            return integer

        # sort by version, descending eg 10.3, 9.4, 8.3
        candidates.sort(key=lambda x: version_to_int(x[2]), reverse=True)

        # print this out for debugging on CI machines if they select verbose
        if self.verbose:
            for root, suffix, version in candidates:
                console(f'Compiler {compiler+suffix} ({version}) at {root+compiler+suffix}')

        root, suffix, version = candidates[0]
        if self.verbose:
            console(f'==> Selected {compiler+suffix} ({version}) at {root+compiler+suffix} <==')
        return root, suffix, version


    def get_preferred_compiler_paths(self):
        if self.cc_path and self.cxx_path and self.cxx_version:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        # no preferred cc path for MSVC and Android
        if self.windows:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        if self.android:
            self.cc_path  = self.android.cc_path()
            self.cxx_path = self.android.cxx_path()
            self.cxx_version = self._get_gcc_clang_fullversion(self.cc_path, dumpfullversion=False)
        elif self.raspi:  # only GCC available for this platform
            ext = '.exe' if System.windows else ''
            self.cc_path  = f'{self.raspi_bin()}arm-linux-gnueabihf-gcc{ext}'
            self.cxx_path = f'{self.raspi_bin()}arm-linux-gnueabihf-g++{ext}'
            self.cxx_version = self._get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.oclea:
            self.cc_path  = f'{self.oclea.bin()}aarch64-oclea-linux-gcc'
            self.cxx_path = f'{self.oclea.bin()}aarch64-oclea-linux-g++'
            self.cxx_version = self._get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.mips:
            self.cc_path  = f'{self.mips.compiler_prefix()}gcc'
            self.cxx_path = f'{self.mips.compiler_prefix()}g++'
            self.cxx_version = self._get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.clang:
            suffixes = ['-12','-11','-10','-9','-8','-7','-6','']
            self.clang_path, suffix, ver = self.find_compiler_root(self.clang_path, 'clang++', suffixes, dumpfullversion=False)
            self.cc_path = f'{self.clang_path}clang{suffix}'
            self.cxx_path = f'{self.clang_path}clang++{suffix}'
            self.cxx_version = ver
        elif self.gcc:
            suffixes = ['-11','-10','-9','-8','-7','-6','']
            self.gcc_path, suffix, ver = self.find_compiler_root(self.gcc_path, 'g++', suffixes, dumpfullversion=True)
            self.cc_path = f'{self.gcc_path}gcc{suffix}'
            self.cxx_path = f'{self.gcc_path}g++{suffix}'
            self.cxx_version = ver

        if self.cc_path and self.cxx_path and self.cxx_version:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        raise EnvironmentError('No preferred compiler for this platform!')


    def _get_gcc_clang_fullversion(self, cc_path, dumpfullversion):
        if dumpfullversion:
            version = execute_piped(f'{cc_path} -dumpfullversion').strip() # eg 9.4.0
            if version.count('.') >= 1:
                return version
        # clang++ doesn't support -dumpfullversion in latest releases -_-
        return execute_piped(f'{cc_path} -dumpversion').strip()


    def compiler_version(self):
        if self.windows:
            msvc_tools = self.get_msvc_tools_path()
            version = os.path.basename(msvc_tools.rstrip('\\//')).split('.')[0]
            return f'msvc{version}'
        elif self.macos:
            return self.macos_version
        elif self.ios:
            return self.ios_version
        elif self.linux or self.raspi or self.oclea or self.mips or self.android:
            cc, _, version = self.get_preferred_compiler_paths()
            version_parts = version.split('.')
            major_version, minor_version = version_parts[0], version_parts[1]
            if 'gcc' in cc: return f'gcc{major_version}.{minor_version}'
            if 'clang' in cc: return f'clang{major_version}.{minor_version}'
            raise EnvironmentError(f'Unrecognized compiler {cc}!')
        else:
            raise EnvironmentError(f'Unknown compiler version!')


    def find_ninja_build(self):
        ninja_executables = [
            os.getenv('NINJA'), 
            util.find_executable_from_system('ninja'),
            '/Projects/ninja.exe'
        ]
        for ninja_exe in ninja_executables:        
            if ninja_exe and os.path.isfile(ninja_exe):
                if self.verbose: console(f'Found Ninja Build System: {ninja_exe}')
                return ninja_exe
        return ''


    def append_env_path(self, paths, env):
        path = os.getenv(env)
        if path: paths.append(path)


    def raspi_bin(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_compilers


    def raspi_sysroot(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_system

    
    def raspi_includes(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_include_paths


    def init_raspi_path(self):
        paths = []
        self.append_env_path(paths, 'RASPI_HOME')
        self.append_env_path(paths, 'RASPBERRY_HOME')
        if System.windows: paths += ['/SysGCC/raspberry']
        elif System.linux: paths += ['/usr/bin/raspberry', '/usr/local/bin/raspberry', '/opt/raspberry']
        compiler = ''
        if System.windows: compiler = 'bin/arm-linux-gnueabihf-gcc.exe'
        elif System.linux: compiler = 'arm-bcm2708/arm-linux-gnueabihf/bin/arm-linux-gnueabihf-gcc'
        for raspi_path in paths:
            if os.path.exists(f'{raspi_path}/{compiler}'):
                if not System.windows:
                    raspi_path = f'{raspi_path}/arm-bcm2708/arm-linux-gnueabihf/'
                self.raspi_compilers = f'{raspi_path}/bin/'
                self.raspi_system    = f'{raspi_path}/arm-linux-gnueabihf/sysroot'
                self.raspi_include_paths = [f'{raspi_path}/arm-linux-gnueabihf/lib/include']
                if self.print: console(f'Found RASPI TOOLS: {self.raspi_compilers}\n    sysroot: {self.raspi_system}')
                return
        raise EnvironmentError(f'''No Raspberry PI toolchain compilers detected! 
Default search paths: {paths} 
Define env RASPI_HOME with path to Raspberry tools.''')


    def set_android_toolchain(self, toolchain_file):
        """
        Sets the toolchain file for Android NDK.
        This should be abspath such as `/opt/android-sdk/ndk/25.2.9519653/build/cmake/android.toolchain.cmake`
        """
        self.android.set_toolchain_path(toolchain_file)


    def set_oclea_toolchain(self, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir where these subdirs exist:
            aarch64-oclea-linux/
            x86_64-ocleasdk-linux/
        And optionally also sets the CMake toolchain file via `toolchain_file`
        """
        self.oclea.init_toolchain(toolchain_dir, toolchain_file)


    def set_mips_toolchain(self, arch, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir for MIPS platform where at least the `bin` dir should exist
        And optionally also sets the CMake toolchain file via `toolchain_file`
        """
        self.mips.init_toolchain(arch, toolchain_dir, toolchain_file)


    def find_default_fortran_compiler(self):
        paths = []
        if System.linux:
            paths += [util.find_executable_from_system('gfortran')]
        
        for fortran_path in paths:
            if fortran_path and os.path.exists(fortran_path):
                if self.verbose: console(f'Found Fortran: {fortran_path}')
                return fortran_path
        return None


    def get_visualstudio_path(self):
        if self._visualstudio_path:
            return self._visualstudio_path
        if not System.windows:
            raise EnvironmentError('VisualStudio tools support not available on this platform!')

        vswhere_exe = "C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\vswhere.exe"
        vspath = execute_piped(f'"{vswhere_exe}" -latest -nologo -property installationPath')
        if vspath and os.path.exists(vspath):
            self._visualstudio_path = vspath
            if self.verbose: console(f'Detected VisualStudio: {vspath}')
            return vspath
        
        paths = []
        vs_variants = [ 'Enterprise', 'Professional', 'Community'  ]
        for version in [ '2022' ]: # new 64-bit VS
            for variant in vs_variants:
                paths.append(f'C:\\Program Files\\Microsoft Visual Studio\\{version}\\{variant}')
        for version in [ '2019', '2017' ]:
            for variant in vs_variants:
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\{version}\\{variant}')

        for path in paths:
            if path and os.path.exists(path):
                self._visualstudio_path = path
                if self.verbose: console(f'Detected VisualStudio: {path}')
                return path

        return self._visualstudio_path


    def is_target_arch_x64(self): return self.arch == 'x64'
    def is_target_arch_x86(self): return self.arch == 'x86'
    def is_target_arch_arm64(self): return self.arch == 'arm64'
    def is_target_arch_armv7(self): return self.arch == 'arm'


    def get_gcc_linux_march(self):
        if self.is_target_arch_x64():
            return 'native' if System.is_64bit else 'x86-64'
        if self.is_target_arch_x86():
            return 'pentium4' if System.is_64bit else 'native'
        raise RuntimeError(f'Unsupported arch: {self.arch}')


    def get_visualstudio_cmake_arch(self):
        if self.is_target_arch_x64(): return 'x64'
        if self.is_target_arch_x86(): return 'Win32'
        if self.arch == 'arm':   return 'ARM'
        if self.arch == 'arm64': return 'ARM64'
        raise RuntimeError(f'Unsupported arch: {self.arch}')


    def get_visualstudio_cmake_id(self):
        if self._visualstudio_cmake_id:
            return self._visualstudio_cmake_id
        
        path = self.get_visualstudio_path()
        if '\\2022\\' in path: self._visualstudio_cmake_id = 'Visual Studio 17 2022'
        elif '\\2019\\' in path: self._visualstudio_cmake_id = 'Visual Studio 16 2019'
        else:                  self._visualstudio_cmake_id = 'Visual Studio 15 2017'
        
        if self.verbose: console(f'Detected CMake Generator: -G"{self._visualstudio_cmake_id}" -A {self.get_visualstudio_cmake_arch()}')
        return self._visualstudio_cmake_id


    def get_msbuild_path(self):
        if self._msbuild_path:
            return self._msbuild_path
        
        paths = [ util.find_executable_from_system('msbuild') ]
        if System.windows:
            vswhere = '"C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\vswhere.exe" -latest -nologo -property installationPath'
            paths.append(f"{execute_piped(vswhere)}\\MSBuild\\Current\\Bin\\MSBuild.exe")
            paths.append(f"{execute_piped(vswhere)}\\MSBuild\\15.0\\Bin\\amd64\\MSBuild.exe")
            
            vs_variants = [ 'Enterprise', 'Professional', 'Community' ]
            for variant in vs_variants:
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\2019\\{variant}\\MSBuild\\Current\\Bin\\MSBuild.exe')
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\2017\\{variant}\\MSBuild\\15.0\\Bin\\amd64\\MSBuild.exe')

        for path in paths:
            if path and os.path.exists(path):
                self._msbuild_path = path
                if self.verbose: console(f'Detected MSBuild: {path}')
                return path
        raise EnvironmentError('Failed to find MSBuild from system PATH. You can easily configure msbuild by running `mama install-msbuild`.')


    ## MSVC tools at, for example: "{VisualStudioPath}\VC\Tools\MSVC\14.16.27023"
    def get_msvc_tools_path(self):
        if self._msvctools_path:
            return self._msvctools_path
        if not System.windows:
            raise EnvironmentError('MSVC tools not available on this platform!')

        tools_root = f"{self.get_visualstudio_path()}\\VC\\Tools\\MSVC"
        tools = os.listdir(tools_root)
        if not tools:
            raise EnvironmentError('Could not detect MSVC Tools')

        tools_path = os.path.join(tools_root, tools[0])
        #tools_path = forward_slashes(tools_path)
        self._msvctools_path = tools_path
        if self.verbose: console(f'Detected MSVC Tools: {tools_path}')
        return tools_path


    def get_msvc_bin64(self):
        return f'{self.get_msvc_tools_path()}/bin/Hostx64/x64/'


    def get_msvc_link64(self):
        return f'{self.get_msvc_bin64()}link.exe'


    def get_msvc_cl64(self):
        return f'{self.get_msvc_bin64()}cl.exe'


    def get_msvc_lib64(self):
        return f'{self.get_msvc_tools_path()}\\lib\\x64'


    def install_clang6(self):
        if System.windows: raise OSError('Install Visual Studio 2019 with Clang support')
        if System.macos:   raise OSError('Install Xcode to get Clang on macOS')

        id, major, _ = self.get_distro_info()
        if id != "ubuntu": raise OSError('install_clang6 only supports ubuntu')
        suffix = '1404'
        if major >= 16: suffix = '1604'

        console(f'Choosing {suffix} for kernel major={major}')
        self.install_clang(clang_major='6', clang_ver='6.0', suffix=suffix)


    def install_clang11(self):
        if System.windows: raise OSError('Install Visual Studio 2019 with Clang support')
        if System.macos:   raise OSError('Install Xcode to get Clang on macOS')

        id, major, minor = self.get_distro_info()
        if id != "ubuntu": raise OSError('install_clang11 only supports ubuntu')
        suffix = '1604'
        if major >= 20 and minor >= 10: suffix = '2010'
        elif major >= 20: suffix = '2004'
        elif major >= 16: suffix = '1604'
        console(f'Choosing {suffix} for kernel major={major} minor={minor}')

        self.install_clang(clang_major='11', clang_ver='11.0', suffix=suffix)


    def install_clang(self, clang_major, clang_ver, suffix):
        clang_major = '11'
        clang_ver = '11.0'
        clangpp = f'clang++{clang_major}'
        clang_zip = util.download_file(f'http://ateh10.net/dev/{clangpp}-{suffix}.zip', tempfile.gettempdir())
        console(f'Installing to /usr/local/{clangpp}')
        execute(f'sudo rm -rf /usr/local/{clangpp}') # get rid of any old stuff
        execute(f'cd /usr/local && sudo unzip -oq {clang_zip}') # extract /usr/local/clang++11/
        os.remove(clang_zip)
        execute(f'sudo ln -sf /usr/local/{clangpp}/lib/libc++.so.1    /usr/lib')
        execute(f'sudo ln -sf /usr/local/{clangpp}/lib/libc++abi.so.1 /usr/lib')
        execute(f'sudo ln -sf /usr/local/{clangpp}/bin/clang      /usr/bin/clang-{clang_ver}')
        execute(f'sudo ln -sf /usr/local/{clangpp}/bin/clang++    /usr/bin/clang++-{clang_ver}')
        execute(f'sudo ln -sf /usr/local/{clangpp}/include/c++/v1 /usr/include/c++/v1')
        execute(f'sudo update-alternatives --install /usr/bin/clang   clang   /usr/bin/clang-{clang_ver}   100')
        execute(f'sudo update-alternatives --install /usr/bin/clang++ clang++ /usr/bin/clang++-{clang_ver} 100')
        execute(f'sudo update-alternatives --set clang   /usr/bin/clang-{clang_ver}')
        execute(f'sudo update-alternatives --set clang++ /usr/bin/clang++-{clang_ver}')


    def install_msbuild(self):
        if System.windows: raise OSError('Install Visual Studio 2019 to get MSBuild on Windows')
        if System.macos:   raise OSError('install_msbuild not implemented for macOS')

        id, _, _ = self.get_distro_info()
        if id != "ubuntu": raise OSError('install_msbuild only supports ubuntu')
        codename = distro.info()['codename']

        execute('curl https://packages.microsoft.com/keys/microsoft.asc | gpg --dearmor > /tmp/microsoft.gpg')
        execute('sudo mv /tmp/microsoft.gpg /etc/apt/trusted.gpg.d/microsoft.gpg')
        execute(f"sudo sh -c 'echo \"deb [arch=amd64] https://packages.microsoft.com/repos/microsoft-ubuntu-{codename}-prod {codename} main\" > /etc/apt/sources.list.d/dotnetdev.list'")
        execute('sudo apt-get install apt-transport-https')
        execute('sudo apt-get update')
        execute('sudo apt-get install dotnet-sdk-2.1')


    def install_ndk(self):
        if System.macos:
            raise OSError('install_ndk not implemented for macOS')
        elif System.windows:
            ndk_version = '25.2.9519653'
            ndk_url = 'https://dl.google.com/android/repository/android-ndk-r25c-windows.zip'
            ndk_dest = f'{os.getenv("LOCALAPPDATA")}\\Android\\Sdk\\ndk'
        elif System.linux:
            ndk_version = '25.2.9519653'
            ndk_url = 'https://dl.google.com/android/repository/android-ndk-r25c-linux.zip'
            ndk_dest = f'/opt/android-sdk/ndk'

        console(f'Downloading NDK {ndk_version}')
        ndk_zip = util.download_file(ndk_url, tempfile.gettempdir())

        if System.windows:
            os.makedirs(ndk_dest, exist_ok=True)
        else:
            execute(f'sudo mkdir -p {ndk_dest} && sudo chown -R $USER {ndk_dest}')

        console(f'Extracting NDK to {ndk_dest}/{ndk_version}')
        util.unzip(ndk_zip, ndk_dest)

        final_dest = f'{ndk_dest}/{ndk_version}'
        if os.path.exists(final_dest):
            shutil.rmtree(final_dest)

        shutil.move(f'{ndk_dest}/android-ndk-r25c', final_dest)
        if os.path.exists(f'{final_dest}/build'):
            console(f'NDK installed successfully to {final_dest}')
        else:
            raise RuntimeError(f'Failed to install NDK to {final_dest}')

        console(f'Adding ANDROID_NDK_HOME={final_dest} to ~/.bashrc, run source ~/.bashrc or restart terminal to populate your env.')
        execute(f'echo "export ANDROID_NDK_HOME={final_dest}" >> ~/.bashrc')


    def run_convenient_installs(self):
        if 'clang6'  in self.convenient_install: self.install_clang6()
        if 'clang11' in self.convenient_install: self.install_clang11()
        if 'msbuild' in self.convenient_install: self.install_msbuild()
        if 'ndk'     in self.convenient_install: self.install_ndk()


    def libname(self, library):
        if self.windows: return f'{library}.lib'
        else:            return f'lib{library}.a'


    def libext(self):
        return 'lib' if self.windows else 'a'


    def targets_all(self) -> bool:
        return self.target == 'all'


    def target_matches(self, target_name: str) -> bool:
        return self.targets_all() \
            or (self.target and self.target.lower() == target_name.lower())


    def no_specific_target(self) -> bool:
        return (not self.target) or self.targets_all()




from __future__ import annotations
from typing import Tuple, TYPE_CHECKING
import os, shlex
from .system import System, console
from .run import get_cwd_exe_args
from .sub_process import execute_echo

if TYPE_CHECKING:
    from ..build_target import BuildTarget


def filter_gdb_arg(args: str, default_gdb=True) -> Tuple[str, bool]:
    if 'nogdb' == args: return '', False
    if 'nogdb ' in args: return args.replace('nogdb ', ''), False
    if 'gdb' == args: return '', True
    if 'gdb ' in args: return args.replace('gdb ', ''), True
    return args, default_gdb


def run_gdb(target: BuildTarget, command: str, src_dir=True):
    if target.android or target.ios or target.raspi or target.oclea or target.mips:
        console('Cannot run tests for Android, iOS, Raspi, Oclea, MIPS builds.')
        return # nothing to run

    root_dir = target.source_dir() if src_dir else target.build_dir()
    if target.windows and not src_dir:
        root_dir = f'{root_dir}/{target.cmake_build_type}'

    cwd, exe, args = get_cwd_exe_args(target, command, root_dir=root_dir)

    if target.windows:
        debugger = f'{exe} {args}'
    elif target.macos:
        # b: batch, q: quiet, -o r: run
        # -k bt: on crash, backtrace
        # -k q: on crash, quit 
        debugger = f'lldb -b -o r -k bt -k q  -- {exe} {args}'
    else: # linux
        # r: run;  bt: give backtrace;  q: quit when done;
        debugger = f'gdb -batch -return-child-result -ex=r -ex=bt -ex=q --args {exe} {args}'

    if not os.path.exists(exe):
        raise IOError(f'Could not find {exe}')
    execute_echo(cwd=cwd, cmd=debugger, exit_on_fail=True)
